/* 
 * tkPack2.c --
 *
 *	This file contains alternative code to implement the "packer"
 *	geometry manager for Tk.
 *
 * Author: Pekka Nikander <pnr@innopoli.ajk.tele.fi>
 *
 * Copyright 1992 Telecom Finland
 * Permission to use, copy, modify, and distribute this
 * software and its documentation for any purpose and without
 * fee is hereby granted.
 *
 * Copyright 1990 Regents of the University of California.
 * Permission to use, copy, modify, and distribute this
 * software and its documentation for any purpose and without
 * fee is hereby granted, provided that the above copyright
 * notice appear in all copies.  The University of California
 * makes no representations about the suitability of this
 * software for any purpose.  It is provided "as is" without
 * express or implied warranty.
 *
 * Created: Wed Mar 11 18:47:48 1992
 * Last modified: Wed Mar 11 21:30:55 1992 pnr
 */

#ifndef lint
static char rcsid[] = "...";
#endif

#include "tkconfig.h"
#include "tkint.h"
#include "tkgeommanager.h"

#define TOP TK_ANCHOR_N
#define LEFT TK_ANCHOR_W
#define RIGHT TK_ANCHOR_E
#define BOTTOM TK_ANCHOR_S


/* For each window that the packer cares about (either because
 * the window is managed by the packer or because the window
 * has children that are managed by the packer), there is a
 * structure of the following type:
 */

typedef struct Packer {
    TkGeomManager *geomPtr;	/* Geometry manager structure this
				 * packer is associated with. */
    Tk_Anchor side;		/* Side of parent against which
				 * this window is packed. */
    Tk_Anchor anchorPoint;	/* If frame allocated for window is larger
				 * than window needs, this indicates how
				 * where to position window in frame. */
    int padX, padY;		/* Additional amounts of space to give window
				 * besides what it asked for. */
    int doubleBw;		/* Twice the window's last known border
				 * width.  If this changes, the window
				 * must be repacked within its parent. */
    int *abortPtr;		/* If non-NULL, it means that there is a nested
				 * call to ArrangePacking already working on
				 * this window.  *abortPtr may be set to 1 to
				 * abort that nested call.  This happens, for
				 * example, if tkwin or any of its children
				 * is deleted. */
    int fillx;			/* 1 means if frame allocated for window
				 * is wider than window needs, expand window
				 * to fill frame.  0 means don't make window
				 * any larger than needed. */
    int filly;			/* Same as FILLX, except for height. */
    int expand;			/* 1 means this window's frame will absorb any
				 * extra space in the parent window. */
    char *previousName;		/* Name of previous window in display
				 * order.  Used by ArrangeChildren. */
} Packer;

/*
 * Hash table used to map from Tk_Window tokens to corresponding
 * Packer structures:
 */

static Tcl_HashTable packerHashTable;

/*
 * Have statics in this module been initialized?
 */

static initialized = 0;

/*
 * Forward declarations for procedures defined later in this file:
 */

static void		ArrangePacking _ANSI_ARGS_((ClientData clientData));
static Packer *		GetPacker _ANSI_ARGS_((Tcl_Interp *interp,
				    Tk_Window tkwin));
static int		PackAfter _ANSI_ARGS_((Tcl_Interp *interp,
			    Packer *parentPtr, int argc, char **argv));
static void		DestroyPacker _ANSI_ARGS_ ((ClientData clientData));


static Tk_ConfigSpec configSpecs[] = {
    {TK_CONFIG_ANCHOR, "-side", "side", "Side",
        "n", Tk_Offset(Packer, side), 0},
    {TK_CONFIG_ANCHOR, "-anchor", "anchor", "Anchor",
        "center", Tk_Offset(Packer, anchorPoint), 0},
    {TK_CONFIG_INT, "-padx", "padX", "PadX",
        "0", Tk_Offset(Packer, padX), 0},
    {TK_CONFIG_INT, "-pady", "padY", "PadY",
        "0", Tk_Offset(Packer, padY), 0},
    {TK_CONFIG_BOOLEAN, "-fillx", "fillX", "FillX",
        "no", Tk_Offset(Packer, fillx), 0},
    {TK_CONFIG_BOOLEAN, "-filly", "fillY", "FillY",
        "no", Tk_Offset(Packer, filly), 0},
    {TK_CONFIG_BOOLEAN, "-expand", "expand", "Expand",
        "no", Tk_Offset(Packer, expand), 0},
    {TK_CONFIG_STRING, "-after", "after", "Widget",
	0, Tk_Offset(Packer, previousName), 0},
    {TK_CONFIG_END, (char *) NULL, (char *) NULL, (char *) NULL,
        (char *) NULL, 0, 0}
};

/*
 *--------------------------------------------------------------
 *
 * Tk_PackCmd --
 *
 *	This procedure is invoked to process the "pack" Tcl command.
 *	See the user documentation for details on what it does.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	See the user documentation.
 *
 *--------------------------------------------------------------
 */

int
Tk_Pack2Cmd(clientData, interp, argc, argv)
    ClientData clientData;	/* Main window associated with
				 * interpreter. */
    Tcl_Interp *interp;		/* Current interpreter. */
    int argc;			/* Number of arguments. */
    char **argv;		/* Argument strings. */
{
    Tk_Window tkwin = (Tk_Window) clientData;
    int c;
    int length; 

    if (argc < 3) {
	Tcl_AppendResult(interp, "wrong # args: should be \"",
		argv[0], " option arg ?arg ...?\"", (char *) NULL);
	return TCL_ERROR;
    }

    c = argv[1][0];
    length = strlen(argv[1]);
    if ((c == 'c') && (strncmp(argv[1], "configure", length) == 0)) {
	Tk_Window tkwin2;
	Packer *packPtr;
	TkGeomManager *geomPtr;

	tkwin2 = Tk_NameToWindow(interp, argv[2], tkwin);
	if (tkwin2 == NULL) {
	    return TCL_ERROR;
	}
	packPtr = GetPacker(interp, tkwin2);
	geomPtr = packPtr->geomPtr;

	if (argc == 3) {
	    return Tk_ConfigureInfo(interp, geomPtr, configSpecs, packPtr,
				    (char *) NULL, 0);
	}
	if (Tk_ConfigureWidget(interp, geomPtr, configSpecs,
			       argc - 3, argv + 3, packPtr, 
			       TK_CONFIG_ARGV_ONLY) != TCL_OK) {
	    return TCL_ERROR;
	}

	if ((geomPtr->parentPtr) != NULL
	        && !(geomPtr->parentPtr->flags & REQUESTED_REARRANGE)) {
	    geomPtr->parentPtr->flags |= REQUESTED_REARRANGE;
	    Tk_DoWhenIdle(geomPtr->parentPtr->arrangeProc,
			  geomPtr->parentPtr->clientData);
	}
    } else if ((c == 'm') && (strncmp(argv[1], "manage", length) == 0)) {
	Packer *packPtr;
	TkGeomManager *parentPtr, *childPtr;
	Tk_Window tkwin2;
	char *prefix = "";

	tkwin2 = Tk_NameToWindow(interp, argv[2], tkwin);
	if (tkwin2 == NULL) {
	    return TCL_ERROR;
	}
	packPtr = GetPacker(interp, tkwin2);
	parentPtr = packPtr->geomPtr;

	if (argc == 3) {
	    for (childPtr = parentPtr->childPtr; childPtr != NULL;
		     childPtr = childPtr->nextPtr) {
		Tcl_AppendResult(interp, prefix, Tk_PathName(childPtr->tkwin),
				 (char *) NULL);
		prefix = " ";
	    }
	    return TCL_OK;
	}
	return PackAfter(interp, parentPtr, argc-3, argv+3);
    } else if ((c == 'u') && (strncmp(argv[1], "unmanage", length) == 0)) {
	Tk_Window tkwin2;
	Packer *packPtr;
	TkGeomManager *geomPtr;

	if (argc != 3) {
	    Tcl_AppendResult(interp, "wrong # args: should be \"",
		    argv[0], " unmanage window\"", (char *) NULL);
	    return TCL_ERROR;
	}
	tkwin2 = Tk_NameToWindow(interp, argv[2], tkwin);
	if (tkwin2 == NULL) {
	    return TCL_ERROR;
	}
	packPtr = GetPacker(interp, tkwin2);
	geomPtr = packPtr->geomPtr;
	if (geomPtr->parentPtr != NULL) {
	    Tk_ManageGeometry(tkwin2, (Tk_GeometryProc *) NULL,
		    (ClientData) NULL);
	    TkUnlinkGeomManager(packPtr);
	    Tk_UnmapWindow(geomPtr->tkwin);
	}
    } else {
	Tcl_AppendResult(interp, "bad option \"", argv[1],
	    "\":  must be configure, manage, or unmanage",
			 (char *) NULL);
	return TCL_ERROR;
    }
    return TCL_OK;
}

/*
 *--------------------------------------------------------------
 *
 * ArrangePacking --
 *
 *	This procedure is invoked (using the Tk_DoWhenIdle
 *	mechanism) to re-layout a set of windows managed by
 *	the packer.  It is invoked at idle time so that a
 *	series of packer requests can be merged into a single
 *	layout operation.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	The packed children of parentPtr may get resized or
 *	moved.
 *
 *--------------------------------------------------------------
 */

static void
ArrangePacking(clientData)
    ClientData clientData;	/* Structure describing parent
				 * whose children are to be
				 * re-layed out. */
{
    register Packer *packPtr = (Packer *) clientData;
    register TkGeomManager *parentPtr = packPtr->geomPtr;
    register TkGeomManager *childPtr;	
    int numExpX, numExpY;	/* # of windows that are expandable in
				 * each direction. */
    int spareX, spareY;		/* Amount of extra space to give to each
				 * expandable window. */
    int leftOverX, leftOverY;	/* Extra chunk of space to give to last
				 * expandable window. */
    int cavityX, cavityY, cavityWidth, cavityHeight;
				/* These variables keep track of the
				 * as-yet-unallocated space remaining in
				 * the middle of the parent window. */
    int frameX, frameY, frameWidth, frameHeight;
				/* These variables keep track of the frame
				 * allocated to the current window. */
    int x, y, width, height;	/* These variables are used to hold the
				 * actual geometry of the current window. */
    int intBWidth;		/* Width of internal border in parent window,
				 * if any. */
    int abort;			/* May get set to non-zero to abort this
				 * repacking operation. */
    int maxWidth, maxHeight, tmp;

    parentPtr->flags &= ~REQUESTED_REARRANGE;

    /*
     * If the parent has no children anymore, then don't do anything
     * at all:  just leave the parent's size as-is.
     */

    if (parentPtr->childPtr == NULL) {
	return;
    }

    /*
     * Abort any nested call to ArrangePacking for this window, since
     * we'll do everything necessary here, and set up so this call
     * can be aborted if necessary.  
     */

    if (packPtr->abortPtr != NULL) {
	*packPtr->abortPtr = 1;
    }
    packPtr->abortPtr = &abort;
    abort = 0;
    Tk_Preserve((ClientData) parentPtr);

    /*
     * Pass #1: scan all the children to figure out the total amount
     * of space needed.  Two separate widths and heights are computed.
     *
     * "Width" and "height" compute the minimum parent size to meet
     * the needs of each window in the direction "where there is
     * flexibility".  For example, if a child is packed TOP, then
     * y is the flexible direction:  the child's requested height
     * will determine its size.  For this window x is the inflexible
     * direction:  the window's width will be determined by the amount
     * of space left in the parent's cavity, not by the window's
     * requested width.  "Width" and "height" are needed in order to
     * compute how much extra space there is, so that it can be divided
     * among the windows that have the EXPAND flag.
     *
     * "MaxWidth" and "maxHeight" compute the minimum parent size to
     * meet all the needs of every window in both directions, flexible
     * or inflexible.  These values are needed to make geometry requests
     * of the parent's parent.
     */

    intBWidth = Tk_InternalBorderWidth(parentPtr->tkwin);
    width = height = maxWidth = maxHeight = 2*intBWidth;
    numExpX = numExpY = 0;
    for (childPtr = parentPtr->childPtr; childPtr != NULL;
	    childPtr = childPtr->nextPtr) {
	packPtr = (Packer *)childPtr->clientData;
	if ((packPtr->side == TOP) || (packPtr->side == BOTTOM)) {
	    tmp = Tk_ReqWidth(childPtr->tkwin) + packPtr->doubleBw
		    + packPtr->padX + width;
	    if (tmp > maxWidth) {
		maxWidth = tmp;
	    }
	    height += Tk_ReqHeight(childPtr->tkwin) + packPtr->doubleBw
		    + packPtr->padY;
	    if (packPtr->expand) {
		numExpY++;
	    }
	} else {
	    tmp = Tk_ReqHeight(childPtr->tkwin) + packPtr->doubleBw
		    + packPtr->padY + height;
	    if (tmp > maxHeight) {
		maxHeight = tmp;
	    }
	    width += Tk_ReqWidth(childPtr->tkwin) + packPtr->doubleBw
		    + packPtr->padX;
	    if (packPtr->expand) {
		numExpX++;
	    }
	}
    }
    if (width > maxWidth) {
	maxWidth = width;
    }
    if (height > maxHeight) {
	maxHeight = height;
    }

    /*
     * If the total amount of space needed in the parent window has
     * changed, then notify the next geometry manager up and requeue
     * ourselves to start again after the parent has had a chance to
     * resize us.
     */

    if ((maxWidth != Tk_ReqWidth(parentPtr->tkwin))
	    || (maxHeight != Tk_ReqHeight(parentPtr->tkwin))) {
	Tk_GeometryRequest(parentPtr->tkwin, maxWidth, maxHeight);
	parentPtr->flags |= REQUESTED_REARRANGE;
	Tk_DoWhenIdle(parentPtr->arrangeProc, parentPtr->clientData);
	goto done;
    }

    /*
     * If there is spare space, figure out how much of it goes to
     * each of the windows that is expandable.
     */

    spareX = Tk_Width(parentPtr->tkwin) - width;
    spareY = Tk_Height(parentPtr->tkwin) - height;
    if ((spareX <= 0) || (numExpX == 0)) {
	leftOverX = 0;
	spareX = 0;
    } else {
	leftOverX = spareX % numExpX;
	spareX /= numExpX;
    }
    if ((spareY <= 0) || (numExpY == 0)) {
	leftOverY = spareY;
	spareY = 0;
    } else {
	leftOverY = spareY % numExpY;
	spareY /= numExpY;
    }

    /*
     * Pass #2: scan the children a second time assigning
     * new sizes.  The "cavity" variables keep track of the
     * unclaimed space in the cavity of the window;  this
     * shrinks inward as we allocate windows around the
     * edges.  The "frame" variables keep track of the space
     * allocated to the current window and its frame.  The
     * current window is then placed somewhere inside the
     * frame, depending on anchorPoint.
     */

    cavityX = cavityY = x = y = intBWidth;
    cavityWidth = Tk_Width(parentPtr->tkwin) - 2*intBWidth;
    cavityHeight = Tk_Height(parentPtr->tkwin) - 2*intBWidth;
    for (childPtr = parentPtr->childPtr; childPtr != NULL;
	    childPtr = childPtr->nextPtr) {
	packPtr = (Packer *)childPtr->clientData;
	if ((packPtr->side == TOP) || (packPtr->side == BOTTOM)) {
	    frameWidth = cavityWidth;
	    frameHeight = Tk_ReqHeight(childPtr->tkwin) + packPtr->padY
		    + packPtr->doubleBw;
	    if (packPtr->expand) {
		frameHeight += spareY;
		numExpY--;
		if (numExpY == 0) {
		    frameHeight += leftOverY;
		}
	    }
	    cavityHeight -= frameHeight;
	    if (cavityHeight < 0) {
		frameHeight += cavityHeight;
		cavityHeight = 0;
	    }
	    frameX = cavityX;
	    if (packPtr->side == TOP) {
		frameY = cavityY;
		cavityY += frameHeight;
	    } else {
		frameY = cavityY + cavityHeight;
	    }
	} else {
	    frameHeight = cavityHeight;
	    frameWidth = Tk_ReqWidth(childPtr->tkwin) + packPtr->padX
		    + packPtr->doubleBw;
	    if (packPtr->expand) {
		frameWidth += spareX;
		numExpX--;
		if (numExpX == 0) {
		    frameWidth += leftOverX;
		}
	    }
	    cavityWidth -= frameWidth;
	    if (cavityWidth < 0) {
		frameWidth += cavityWidth;
		cavityWidth = 0;
	    }
	    frameY = cavityY;
	    if (packPtr->side == LEFT) {
		frameX = cavityX;
		cavityX += frameWidth;
	    } else {
		frameX = cavityX + cavityWidth;
	    }
	}

	/*
	 * Now that we've got the size of the frame for the window,
	 * compute the window's actual size and location using the
	 * fill and frame factors.
	 */

#ifndef TK_PACK_PADDIFF
	width = Tk_ReqWidth(childPtr->tkwin) + packPtr->doubleBw;
	if ((packPtr->fillx) || (width > frameWidth)) {
	    width = frameWidth;
	}
	height = Tk_ReqHeight(childPtr->tkwin) + packPtr->doubleBw;
	if ((packPtr->filly) || (height > frameHeight)) {
	    height = frameHeight;
	}
#else
	width = Tk_ReqWidth(childPtr->tkwin) + packPtr->doubleBw;
	if (width > frameWidth) {
	    width = frameWidth;
	} else if (packPtr->fillx) {
	    width = frameWidth - packPtr->padX;
	}
	height = Tk_ReqHeight(childPtr->tkwin) + packPtr->doubleBw;
	if (height > frameHeight) {
	    height = frameHeight;
	} else if (packPtr->filly) {
	    height = frameHeight - packPtr->padY;
	}
#endif
	switch (packPtr->anchorPoint) {
	    case TK_ANCHOR_N:
		x = frameX + (frameWidth - width)/2;
		y = frameY;
		break;
	    case TK_ANCHOR_NE:
		x = frameX + frameWidth - width;
		y = frameY;
		break;
	    case TK_ANCHOR_E:
		x = frameX + frameWidth - width;
		y = frameY + (frameHeight - height)/2;
		break;
	    case TK_ANCHOR_SE:
		x = frameX + frameWidth - width;
		y = frameY + frameHeight - height;
		break;
	    case TK_ANCHOR_S:
		x = frameX + (frameWidth - width)/2;
		y = frameY + frameHeight - height;
		break;
	    case TK_ANCHOR_SW:
		x = frameX;
		y = frameY + frameHeight - height;
		break;
	    case TK_ANCHOR_W:
		x = frameX;
		y = frameY + (frameHeight - height)/2;
		break;
	    case TK_ANCHOR_NW:
		x = frameX;
		y = frameY;
		break;
	    case TK_ANCHOR_CENTER:
		x = frameX + (frameWidth - width)/2;
		y = frameY + (frameHeight - height)/2;
		break;
	    default:
		panic("bad frame factor in ArrangePacking");
	}
	width -= packPtr->doubleBw;
	height -= packPtr->doubleBw;

	/*
	 * If the window is too small to be interesting then
	 * unmap it.  Otherwise configure it and then make sure
	 * it's mapped.
	 */

	if ((width <= 0) || (height <= 0)) {
	    Tk_UnmapWindow(childPtr->tkwin);
	} else {
	    if ((x != Tk_X(childPtr->tkwin))
		    || (y != Tk_Y(childPtr->tkwin))
		    || (width != Tk_Width(childPtr->tkwin))
		    || (height != Tk_Height(childPtr->tkwin))) {
		Tk_MoveResizeWindow(childPtr->tkwin, x, y,
			(unsigned int) width, (unsigned int) height);
	    }
	    if (abort) {
		goto done;
	    }
	    Tk_MapWindow(childPtr->tkwin);
	}

	/*
	 * Changes to the window's structure could cause almost anything
	 * to happen, including deleting the parent or child.  If this
	 * happens, we'll be told to abort.
	 */

	if (abort) {
	    goto done;
	}
    }

    done:
    packPtr = (Packer *)parentPtr->clientData;
    packPtr->abortPtr = NULL;
    Tk_Release((ClientData) parentPtr);
}

/*
 *--------------------------------------------------------------
 *
 * GetPacker --
 *
 *	This internal procedure is used to locate a Packer
 *	structure for a given window, creating one if one
 *	doesn't exist already.
 *
 * Results:
 *	The return value is a pointer to the Packer structure
 *	corresponding to tkwin.
 *
 * Side effects:
 *	A new packer structure may be created.  If so, then
 *	a callback is set up to clean things up when the
 *	window is deleted.
 *
 *--------------------------------------------------------------
 */

static Packer *
GetPacker(interp, tkwin)
    Tcl_Interp *interp;
    Tk_Window tkwin;		/* Token for window for which
				 * packer structure is desired. */
{
    register Packer *packPtr;
    Tcl_HashEntry *hPtr;
    int new;

    if (!initialized) {
	initialized = 1;
	Tcl_InitHashTable(&packerHashTable, TCL_ONE_WORD_KEYS);
    }

    /*
     * See if there's already packer for this window.  If not,
     * then create a new one.
     */

    hPtr = Tcl_CreateHashEntry(&packerHashTable, (char *) tkwin, &new);
    if (!new) {
	return (Packer *) Tcl_GetHashValue(hPtr);
    }

    packPtr = (Packer *) ckalloc(sizeof(Packer));
    packPtr->geomPtr = TkCreateGeomManager(interp, tkwin, "pack",
			   ArrangePacking, DestroyPacker, (ClientData)packPtr);
    packPtr->side = TOP;
    packPtr->anchorPoint = TK_ANCHOR_CENTER;
    packPtr->padX = packPtr->padY = 0;
    packPtr->doubleBw = 2*Tk_Changes(tkwin)->border_width;
    packPtr->abortPtr = NULL;
    packPtr->fillx = packPtr->filly = packPtr->expand = 0;
    packPtr->previousName = NULL;
    Tcl_SetHashValue(hPtr, packPtr);

    Tk_ConfigureWidget(interp, packPtr->geomPtr, configSpecs,
		       0, NULL, packPtr, 0);

    return packPtr;
}

/*
 *--------------------------------------------------------------
 *
 * PackAfter --
 *
 *	This procedure does most of the real work of adding
 *	one or more windows into the packing order for its parent.
 *
 * Results:
 *	A standard Tcl return value.
 *
 * Side effects:
 *	The geometry of the specified windows may change, both now and
 *	again in the future.
 *
 *--------------------------------------------------------------
 */

static int
PackAfter(interp, parentPtr, argc, argv)
    Tcl_Interp *interp;		/* Interpreter for error reporting. */
    TkGeomManager *parentPtr;
				/* Parent in which to pack windows. */
    int argc;			/* Number of elements in argv. */
    char **argv;		/* Array of lists, each containing 2
				 * elements:  window name and side
				 * against which to pack. */
{
    TkGeomManager *prevPtr;
    register Packer *packPtr;
    Tk_Window tkwin;

    /*
     * By default add to the end of window order.
     */
    prevPtr = parentPtr->childPtr;
    if (prevPtr != NULL) {
	while (prevPtr->nextPtr != NULL) {
	    prevPtr = prevPtr->nextPtr;
	}
    }

    /*
     * Iterate over all of the window specifiers.
     */

    for ( ; argc > 0; argc--, argv++, prevPtr = packPtr->geomPtr) {

	/*
	 * Find the packer for the window to be packed, and make sure
	 * that the window in which it will be packed is its parent.
	 */

	tkwin = Tk_NameToWindow(interp, argv[0], parentPtr->tkwin);
	if (tkwin == NULL) {
	    return TCL_ERROR;
	}
	if (Tk_Parent(tkwin) != parentPtr->tkwin) {
	    Tcl_AppendResult(interp, "tried to pack \"",
		    argv[0], "\" in window that isn't its parent",
		    (char *) NULL);
	    return TCL_ERROR;
	}
	packPtr = GetPacker(interp, tkwin);

	/*
	 * Add the packer to the managed list of parent window's
	 * geometry manager.
	 */
	TkLinkGeomManager(parentPtr, packPtr->geomPtr, prevPtr);
    }
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * DestroyPacker --
 *
 *	This procedure is invoked by TkDestroyGeomManager to clean up
 *	the internal structure of a packer at a safe time (when no-one
 *	is using it anymore).
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Everything associated with the packer is freed up.
 *
 *----------------------------------------------------------------------
 */

static void
DestroyPacker(clientData)
    ClientData clientData;		/* Info about packed window that
					 * is now dead. */
{
    register Packer *packPtr = (Packer *) clientData;
    ckfree((char *) packPtr);
}
