/*
 * tkPorthole.c --
 *
 *	This file contains code to implement the "porthole" or "xy"
 *	geometry manager for Tk.
 *
 * Author: Pekka Nikander <pnr@ajk.tele.fi>
 *
 * Copyright 1992 Telecom Finland
 *
 * Permission to use, copy, modify, and distribute this
 * software and its documentation for any purpose and without
 * fee is hereby granted, provided that the above copyright
 * notice appear in all copies.  Telecom Finland
 * makes no representations about the suitability of this
 * software for any purpose.  It is provided "as is" without
 * express or implied warranty.
 *
 * Created: Tue Jan 21 13:33:17 1992
 * Last modified: Wed Mar 11 22:34:23 1992 pnr
 */

#ifndef lint
static char rcsid[] = "...";
#endif

#include <assert.h>
#include <string.h> 

#include "tkconfig.h"
#include "tkint.h"
#include "tkgeommanager.h"

/*
 * For each window that the porthole cares about, there is
 * a structure of the following type:
 */

typedef struct Porthole {
    TkGeomManager *geomPtr;
    int newX, newY;		/* New location requested */
    char *scrollXCmd;		/* Command prefix for cummunication with
				 * X scrollbar(s).  NULL means no command
				 * to issue.  Malloc'ed. */
    char *scrollYCmd;		/* Ditto for Y scrollbar(s). */
    int shown;			/* Shown or not. */
} Porthole;

/*
 * Hash table used to map from Tk_Window tokens to corresponding
 * Porthole structures:
 */

static Tcl_HashTable portholeHashTable;

/*
 * Have statics in this module been initialized?
 */

static initialized = 0;

/* 
 * Forward declarations for procedures defined later in this file:
 */

static void	      ArrangePorthole _ANSI_ARGS_((ClientData clientData));
static Porthole *     GetPorthole _ANSI_ARGS_((Tk_Interp *interp,
					       Tk_Window tkwin));
static void	      PortholeUpdateScrollbars _ANSI_ARGS_((Porthole *phole));
static int	      PortholeAfter _ANSI_ARGS_((Tcl_Interp *interp,
			    Porthole *parentPtr, int argc, char **argv));
static void	      DestroyPorthole _ANSI_ARGS_ ((ClientData clientData));


static Tk_ConfigSpec configSpecs[] = {
    {TK_CONFIG_STRING, "-scrollxcommand", "scrollXCommand", "ScrollCommand",
        0, Tk_Offset(Porthole, scrollXCmd), 0},
    {TK_CONFIG_STRING, "-scrollycommand", "scrollYCommand", "ScrollCommand",
        0, Tk_Offset(Porthole, scrollYCmd), 0},
    {TK_CONFIG_BOOLEAN, "-shown", "shown", "Shown",
	"yes", Tk_Offset(Porthole, shown), 0},
    {TK_CONFIG_INT, "-x", "x", "X", 
        0, Tk_Offset(Porthole, newX), 0},
    {TK_CONFIG_INT, "-y", "y", "Y",
        0, Tk_Offset(Porthole, newY), 0},
    {TK_CONFIG_END, (char *) NULL, (char *) NULL, (char *) NULL,
        (char *) NULL, 0, 0}
};



/*
 *----------------------------------------------------------------
 *
 *
 * Tk_PortholeCmd --
 *
 *	This procedure is invoked to process the "porthole" Tcl
 *	command.  See the user documentation for detais on what
 *	it does. 
 *
 * Results:
 *	A standard Tcl return value.
 *
 * Side effects:
 *	See the user documentation.
 */

int
Tk_PortholeCmd(clientData, interp, argc, argv)
    ClientData clientData;	/* Main window associated with
				 * interpreter. */
    Tcl_Interp *interp;		/* Current interpreter. */
    int argc;			/* Number of arguments. */
    char **argv;		/* Argument strings. */
{
    Tk_Window tkwin = (Tk_Window) clientData;
    int c;
    int length; 

    if (argc < 3) {
	Tcl_AppendResult(interp, "wrong # args: should be \"",
		argv[0], " option arg ?arg ...?\"", (char *) NULL);
	return TCL_ERROR;
    }

    c = argv[1][0];
    length = strlen(argv[1]);
    if ((c == 'c') && (strncmp(argv[1], "configure", length) == 0)) {
	Tk_Window tkwin2;
	Porthole *portPtr;
	TkGeomManager *geomPtr;

	tkwin2 = Tk_NameToWindow(interp, argv[2], tkwin);
	if (tkwin2 == NULL) {
	    return TCL_ERROR;
	}
	portPtr = GetPorthole(interp, tkwin2);
	geomPtr = portPtr->geomPtr;

	if (argc == 3) {
	    return Tk_ConfigureInfo(interp, geomPtr, configSpecs, portPtr,
				    (char *) NULL, 0);
	}
	if (Tk_ConfigureWidget(interp, geomPtr, configSpecs,
			       argc - 3, argv + 3, portPtr, 
			       TK_CONFIG_ARGV_ONLY) != TCL_OK) {
	    return TCL_ERROR;
	}

	if ((geomPtr->parentPtr) != NULL
	        && !(geomPtr->parentPtr->flags & REQUESTED_REARRANGE)) {
	    geomPtr->parentPtr->flags |= REQUESTED_REARRANGE;
	    Tk_DoWhenIdle(geomPtr->parentPtr->arrangeProc,
			  geomPtr->parentPtr->clientData);
	}
    } else if ((c == 'm') && (strncmp(argv[1], "manage", length) == 0)) {
	Porthole *portPtr;
	TkGeomManager *parentPtr, *childPtr;
	Tk_Window tkwin2;
	char *prefix = "";

	tkwin2 = Tk_NameToWindow(interp, argv[2], tkwin);
	if (tkwin2 == NULL) {
	    return TCL_ERROR;
	}
	portPtr = GetPorthole(interp, tkwin2);
	parentPtr = portPtr->geomPtr;

	if (argc == 3) {
	    for (childPtr = parentPtr->childPtr; childPtr != NULL;
		     childPtr = childPtr->nextPtr) {
		Tcl_AppendResult(interp, prefix, Tk_PathName(childPtr->tkwin),
				 (char *) NULL);
		prefix = " ";
	    }
	    return TCL_OK;
	}
	return PortAfter(interp, parentPtr, argc-3, argv+3);
    } else if ((c == 'u') && (strncmp(argv[1], "unmanage", length) == 0)) {
	Tk_Window tkwin2;
	Porthole *portPtr;
	TkGeomManager *geomPtr;

	if (argc != 3) {
	    Tcl_AppendResult(interp, "wrong # args: should be \"",
		    argv[0], " unmanage window\"", (char *) NULL);
	    return TCL_ERROR;
	}
	tkwin2 = Tk_NameToWindow(interp, argv[2], tkwin);
	if (tkwin2 == NULL) {
	    return TCL_ERROR;
	}
	portPtr = GetPorthole(interp, tkwin2);
	geomPtr = portPtr->geomPtr;
	if (geomPtr->parentPtr != NULL) {
	    Tk_ManageGeometry(tkwin2, (Tk_GeometryProc *) NULL,
		    (ClientData) NULL);
	    TkUnlinkGeomManager(portPtr);
	    Tk_UnmapWindow(geomPtr->tkwin);
	}
    } else {
	Tcl_AppendResult(interp, "bad option \"", argv[1],
	    "\":  must be configure, manage, or unmanage",
			 (char *) NULL);
	return TCL_ERROR;
    }
    return TCL_OK;
}

/*
 *--------------------------------------------------------------
 *
 * ArrangePorthole --
 *
 *	This procedure is invoked (using the Tk_DoWhenIdle
 *	mechanism) to re-layout the window(s) currently shown by
 *	the porthole.  It is invoked at idle time so that a
 *	series of porthole requests can be merged into a single
 *	operation.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	The window positions are changed if requested.
 *
 *--------------------------------------------------------------
 */

static void
ArrangePorthole(clientData)
    ClientData clientData;	/* Structure describing parent
				 * whose children are to be
				 * re-layed out. */
{
    register Porthole *portPtr = (Porthole *) clientData;
    register TkGeomManager *parentPtr = portPtr->geomPtr;
    register TkGeomManager *childPtr;
    Tk_Window parent, child;

    parentPtr->flags &= ~REQUESTED_REARRANGE;
    Tk_Preserve(parentPtr);
    parent = parentPtr->tkwin;

    for (childPtr = parentPtr->childPtr;
	 childPtr != NULL;
	 childPtr = childPtr->nextPtr) {

	Tk_Preserve(childPtr);
	child = childPtr->tkwin;
	portPtr = (Porthole *)childPtr->clientData;

	/* 
	 * If the window shown is larger than
	 * the porthole, we take the origin
	 * as negative, which leads to natural
	 * behaviour with scrollbars.
	 *
	 * NOTE!  If the size of window shown
	 * 	  is smaller than the porthole, we
	 * 	  allow the user to move the window around... 
	 *
	 * NOTE!  If the child is larger than the parent,
	 *	  and we suddenly change x or y to the
	 *	  opposite of the value (e.g. from -10 to 10),
	 *	  this will fail...
	 */
	
	if (portPtr->newX != Tk_X(child)
	    && Tk_ReqWidth(child)  > Tk_Width(parent)) {
	    portPtr->newX = -(portPtr->newX);
	}
	if (portPtr->newY != Tk_Y(child)
	    && Tk_ReqHeight(child) > Tk_Height(parent)) {
	    portPtr->newY = -(portPtr->newY);
	}
	
	if (!(portPtr->shown)
	       || Tk_ReqWidth(child) == 0
	       || Tk_ReqHeight(child) == 0) {
	    Tk_UnmapWindow(child);
	}
	/*
	 * If the location or size have changed,
	 * move and resize.
	 */
	if ((Tk_ReqWidth(child) > 0 && Tk_ReqHeight(child) > 0)
	    &&(   Tk_ReqWidth(child)  != Tk_Width(child)
	       || Tk_ReqHeight(child) != Tk_Height(child)
	       || portPtr->newX != Tk_X(child)
	       || portPtr->newY != Tk_Y(child))) {
	    Tk_MoveResizeWindow(child, portPtr->newX, portPtr->newY,
				Tk_ReqWidth(child), Tk_ReqHeight(child));
	}
	
	/*
	 * Map or unmap this window if needed.
	 */
	if ((portPtr->shown) 
	      && (Tk_ReqWidth(child) > 0)
	      && (Tk_ReqHeight(child) > 0)) {
	    Tk_MapWindow(child);
	}
	
	/*
	 * Update scrollbars, if any
	 */
	PortholeUpdateScrollbars(portPtr);

	Tk_Release(childPtr);
    }
    Tk_Release(parentPtr);
}

/*
 *----------------------------------------------------------------------
 *
 * PortholeUpdateScrollbars --
 *
 *	This procedure is invoked whenever information has changed in
 *	a porthole in a way that would invalidate a scrollbar display.
 *	If there are associated scrollbar(s), then this command updates
 *	it by invoking a Tcl command.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	A Tcl command is invoked, and an additional command may be
 *	invoked to process errors in the command.
 *
 *----------------------------------------------------------------------
 */

static void
PortholeUpdateScrollbars(portPtr)
    register Porthole *portPtr;		/* Information about porthole. */
{
    char string[60];
    int result;
    TkGeomManager *geomPtr = portPtr->geomPtr;
    TkGeomManager *parentPtr = geomPtr->parentPtr;

    int total_width, total_height;
    int window_width, window_height;
    int start_x, start_y;
    int stop_x, stop_y;

    assert(parentPtr != NULL);

    if (portPtr->scrollXCmd != NULL) {
	if (!(portPtr->shown)) {
	    start_x = 0;
	    total_width = window_width = stop_x = Tk_Width(parentPtr->tkwin);
	} else {
	    if (Tk_Width(geomPtr->tkwin) > Tk_Width(parentPtr->tkwin)) {
		total_width  = Tk_Width(geomPtr->tkwin);
		window_width = Tk_Width(parentPtr->tkwin);
		start_x	= -(Tk_X(geomPtr->tkwin));
		stop_x	= -(Tk_X(geomPtr->tkwin)) + Tk_Width(parentPtr->tkwin);
	    } else {
		total_width  = Tk_Width(parentPtr->tkwin);
		window_width = Tk_Width(geomPtr->tkwin);
		start_x	= Tk_X(geomPtr->tkwin);
		stop_x 	= Tk_X(geomPtr->tkwin) + Tk_Width(geomPtr->tkwin);
	    }
	}
	
	if (stop_x  >= total_width) {
	    stop_x  = total_width - 1;
	    start_x = stop_x - window_width;
	}
	if (stop_x  <  0)	    stop_x  = 0;
	if (start_x <  0) 	    start_x = 0;

	sprintf(string, " %d %d %d %d",
		total_width, window_width, start_x, stop_x);
	result = Tcl_VarEval(geomPtr->mainPtr->interp,
			     portPtr->scrollXCmd, string,
			     (char *) NULL);
	if (result != TCL_OK) {
	    Tk_BackgroundError(geomPtr->mainPtr->interp);
	}
    }

    if (portPtr->scrollYCmd != NULL) {
	if (!(portPtr->shown)) {
	    start_y = 0;
	    total_height = window_height = stop_y
	      = Tk_Height(parentPtr->tkwin);
	} else {
	    if (Tk_Height(geomPtr->tkwin) > Tk_Height(parentPtr->tkwin)) {
		total_height  = Tk_Height(geomPtr->tkwin);
		window_height = Tk_Height(parentPtr->tkwin);
		start_y	= -(Tk_Y(geomPtr->tkwin));
		stop_y	= -(Tk_Y(geomPtr->tkwin))+ Tk_Height(parentPtr->tkwin);
	    } else {
		total_height  = Tk_Height(parentPtr->tkwin);
		window_height = Tk_Height(geomPtr->tkwin);
		start_y	= Tk_Y(geomPtr->tkwin);
		stop_y 	= Tk_Y(geomPtr->tkwin) + Tk_Height(geomPtr->tkwin);
	    }
	}
	
	if (stop_y  >= total_height) {
	    stop_y  = total_height - 1;
	    start_y = stop_y - window_height;
	}
	if (stop_y  <  0)		 stop_y  = 0;
	if (start_y <  0) 		 start_y = 0;
	
	sprintf(string, " %d %d %d %d",
		total_height, window_height, start_y, stop_y);
	result = Tcl_VarEval(geomPtr->mainPtr->interp,
			     portPtr->scrollYCmd, string,
			     (char *) NULL);
	if (result != TCL_OK) {
	    Tk_BackgroundError(geomPtr->mainPtr->interp);
	}
    }
}

/*
 *--------------------------------------------------------------
 *
 * GetPorthole --
 *
 *	This internal procedure is used to locate a Porthole
 *	structure for a given window, creating one if one
 *	doesn't exist already and is needed.
 *
 * Results:
 *	The return value is a pointer to the Porthole structure
 *	corresponding to tkwin.
 *
 * Side effects:
 *	A new porthole structure may be created.  If so, then
 *	a callback is set up to clean things up when the
 *	window is deleted.
 *
 *--------------------------------------------------------------
 */

static Porthole *
GetPorthole(interp, tkwin)
    Tcl_Interp *interp;
    Tk_Window tkwin;		/* Token for window for which
				 * porthole structure is desired. */
{
    register Porthole *portPtr;
    Tcl_HashEntry *hPtr;
    int new;

    if (!initialized) {
	initialized = 1;
	Tcl_InitHashTable(&portholeHashTable, TCL_ONE_WORD_KEYS);
    }

    /*
     * See if there's already porthole for this window.  If not,
     * then create a new one.
     */

    hPtr = Tcl_CreateHashEntry(&portholeHashTable, (char *) tkwin, &new);
    if (!new) {
	return (Porthole *) Tcl_GetHashValue(hPtr);
    }

    portPtr = (Porthole *) ckalloc(sizeof(Porthole));
    portPtr->geomPtr = TkCreateGeomManager(interp, tkwin, "porthole",
			   ArrangePorthole, DestroyPorthole,
			   (ClientData)portPtr);
    portPtr->newX = portPtr->newY = 0;
    portPtr->scrollXCmd = portPtr->scrollYCmd = NULL;
    portPtr->shown = 1;
    Tcl_SetHashValue(hPtr, portPtr);

    Tk_ConfigureWidget(interp, portPtr->geomPtr, configSpecs,
		       0, NULL, portPtr, 0);

    return portPtr;
}

/*
 *--------------------------------------------------------------
 *
 * PortAfter --
 *
 *	This procedure does most of the real work of adding
 *	one or more windows into the shown set of it's parent.
 *
 * Results:
 *	A standard Tcl return value.
 *
 * Side effects:
 *	The geometry of the specified windows may change, both now and
 *	again in the future.
 *
 *--------------------------------------------------------------
 */

static int
PortAfter(interp, parentPtr, argc, argv)
    Tcl_Interp *interp;		/* Interpreter for error reporting. */
    TkGeomManager *parentPtr;
				/* Parent in which to port windows. */
    int argc;			/* Number of elements in argv. */
    char **argv;		/* Array of window names. */
{
    TkGeomManager *prevPtr;
    register Porthole *portPtr;
    Tk_Window tkwin;

    /*
     * By default add to the end of window order.
     */
    prevPtr = parentPtr->childPtr;
    if (prevPtr != NULL) {
	while (prevPtr->nextPtr != NULL) {
	    prevPtr = prevPtr->nextPtr;
	}
    }

    /*
     * Iterate over all of the window specifiers.
     */

    for ( ; argc > 0; argc--, argv++, prevPtr = portPtr->geomPtr) {

	/*
	 * Find the porthole for the window to be ported, and make sure
	 * that the window in which it will be ported is its parent.
	 */

	tkwin = Tk_NameToWindow(interp, argv[0], parentPtr->tkwin);
	if (tkwin == NULL) {
	    return TCL_ERROR;
	}
	if (Tk_Parent(tkwin) != parentPtr->tkwin) {
	    Tcl_AppendResult(interp, "tried to port \"",
		    argv[0], "\" in window that isn't its parent",
		    (char *) NULL);
	    return TCL_ERROR;
	}
	portPtr = GetPorthole(interp, tkwin);

	/*
	 * Add the porthole to the managed list of parent window's
	 * geometry manager.
	 */
	TkLinkGeomManager(parentPtr, portPtr->geomPtr, prevPtr);
    }
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * DestroyPorthole --
 *
 *	This procedure is invoked by TkDestroyGeomManager to clean up
 *	the internal structure of a porthole at a safe time (when no-one
 *	is using it anymore).
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Everything associated with the porthole is freed up.
 *
 *----------------------------------------------------------------------
 */

static void
DestroyPorthole(clientData)
    ClientData clientData;		/* Info about ported window that
					 * is now dead. */
{
    register Porthole *portPtr = (Porthole *) clientData;
    ckfree((char *) portPtr);
}
