/*
 * tkTable.c --
 *
 *	This file contains code to implement the "table"
 *	geometry manager for Tk.
 *
 * Author: Pekka Nikander <pnr@innopoli.ajk.tele.fi>
 *
 * Copyright 1992 Telecom Finland
 *
 * Permission to use, copy, modify, and distribute this
 * software and its documentation for any purpose and without
 * fee is hereby granted, provided that the above copyright
 * notice appear in all copies.  Telecom Finland
 * makes no representations about the suitability of this
 * software for any purpose.  It is provided "as is" without
 * express or implied warranty.
 *
 *
 * Created: Mon Jan 27 12:49:41 1992
 * Last modified: Wed Mar 11 22:34:36 1992 pnr
 */

#ifndef lint
static char rcsid[] = "...";
#endif

#include <assert.h>
#include <math.h>

#include "tkconfig.h"
#include "tkint.h"
#include "tkgeommanager.h"

/*
 * For each window that the table cares about (either because
 * the window is managed by the table or because the window
 * has children that are managed by the table), there is a
 * structure of the following type:
 */

typedef struct Table {
    TkGeomManager *geomPtr;
    int rowNumber;		/* Number of rows */
    int columnNumber;		/* Number of columns */
    int rowHeight;		/* Granted height for a row. */
    int columnWidth;		/* Granted width for a column. */

    int row, col;		/* Current location in table */
    Tk_Anchor anchorPoint;	/* If frame allocated for window is larger
				 * than window needs, this indicates how
				 * where to position window in frame. */
    int fillx;			/* 1 means frame allocated for window
				 *   is wider than window needs, expand 
				 *   window to fill frame.
				 * 0 means don't make window
				 *   any larger than needed. */
    int filly;			/* Same as fillx, except for height. */
    int arrange_vertical;	/* 1 means place second below first etc.
				 *   until rowNumber is reached, then
				 *   start new column.
				 * 0 means place second right of first etc.
				 *   until columnNumber is reached, then
				 *   start new row. */
    int arrange_square;		/* 1 means try to make column and row
				 *   number as equal as possible.
				 * 0 means touch only columnNumber if
				 *   arrange_vertical = 1, or rowNumber
				 *   if arrange_vertical = 0.
				 */
    char *previousName;		/* Name of previous widget, to set
				 * them into correct order. */
} Table;

/*
 * Hash table used to map from Tk_Window tokens to corresponding
 * Table structures:
 */

static Tcl_HashTable tableHashTable;

/*
 * Have statics in this module been initialized?
 */

static initialized = 0;

/*
 * Forward declarations for procedures defined later in this file:
 */

static void		ArrangeTable _ANSI_ARGS_((ClientData clientData));
static Table *		GetTable _ANSI_ARGS_((Tcl_Interp *interp,
					      Tk_Window tkwin));
static void		PlaceChild _ANSI_ARGS_((Table *child, Table *parent));
static int	      	TableAfter _ANSI_ARGS_((Tcl_Interp *interp,
			    Table *parentPtr, int argc, char **argv));
static void	      	DestroyTable _ANSI_ARGS_ ((ClientData clientData));

static Tk_ConfigSpec configSpecs[] = {
    {TK_CONFIG_STRING, "-after", "after", "Widget",
	0, Tk_Offset(Table, previousName), 0},
    {TK_CONFIG_INT, "-columns", "columns", "Columns", 
	"1", Tk_Offset(Table, columnNumber), 0},
    {TK_CONFIG_BOOLEAN, "-fillx", "fillX", "Fill",
	"yes", Tk_Offset(Table, fillx), 0},
    {TK_CONFIG_BOOLEAN, "-filly", "fillY", "Fill",
	"yes", Tk_Offset(Table, filly), 0},
    {TK_CONFIG_ANCHOR, "-frame", "frame", "Frame",
        "w", Tk_Offset(Table, anchorPoint), 0},
    {TK_CONFIG_INT, "-rows", "rows", "Rows",
        "1", Tk_Offset(Table, rowNumber), 0},
    {TK_CONFIG_BOOLEAN, "-square", "square", "Square",
        "no", Tk_Offset(Table, arrange_square), 0},
    {TK_CONFIG_BOOLEAN, "-vertical", "vertical", "Vertical",
        "no", Tk_Offset(Table, arrange_vertical), 0},
    {TK_CONFIG_END, (char *) NULL, (char *) NULL, (char *) NULL,
        (char *) NULL, 0, 0}
};


/*
 *--------------------------------------------------------------
 *
 * Tk_TableCmd --
 *
 *	This procedure is invoked to process the "table" Tcl command.
 *	See the user documentation for details on what it does.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	See the user documentation.
 *
 *--------------------------------------------------------------
 */

int
Tk_TableCmd(clientData, interp, argc, argv)
    ClientData clientData;	/* Main window associated with
				 * interpreter. */
    Tcl_Interp *interp;		/* Current interpreter. */
    int argc;			/* Number of arguments. */
    char **argv;		/* Argument strings. */
{
    Tk_Window tkwin = (Tk_Window) clientData;
    int c;
    int length; 

    if (argc < 3) {
	Tcl_AppendResult(interp, "wrong # args: should be \"",
		argv[0], " option arg ?arg ...?\"", (char *) NULL);
	return TCL_ERROR;
    }

    c = argv[1][0];
    length = strlen(argv[1]);
    if ((c == 'c') && (strncmp(argv[1], "configure", length) == 0)) {
	Tk_Window tkwin2;
	Table *tablePtr;
	TkGeomManager *geomPtr;

	tkwin2 = Tk_NameToWindow(interp, argv[2], tkwin);
	if (tkwin2 == NULL) {
	    return TCL_ERROR;
	}
	tablePtr = GetTable(interp, tkwin2);
	geomPtr = tablePtr->geomPtr;

	if (argc == 3) {
	    return Tk_ConfigureInfo(interp, geomPtr, configSpecs, tablePtr,
				    (char *) NULL, 0);
	}
	if (Tk_ConfigureWidget(interp, geomPtr, configSpecs,
			       argc - 3, argv + 3, tablePtr, 
			       TK_CONFIG_ARGV_ONLY) != TCL_OK) {
	    return TCL_ERROR;
	}

	if ((geomPtr->parentPtr) != NULL
	        && !(geomPtr->parentPtr->flags & REQUESTED_REARRANGE)) {
	    geomPtr->parentPtr->flags |= REQUESTED_REARRANGE;
	    Tk_DoWhenIdle(geomPtr->parentPtr->arrangeProc,
			  geomPtr->parentPtr->clientData);
	}
    } else if ((c == 'm') && (strncmp(argv[1], "manage", length) == 0)) {
	Table *tablePtr;
	TkGeomManager *parentPtr, *childPtr;
	Tk_Window tkwin2;
	char *prefix = "";

	tkwin2 = Tk_NameToWindow(interp, argv[2], tkwin);
	if (tkwin2 == NULL) {
	    return TCL_ERROR;
	}
	tablePtr = GetTable(interp, tkwin2);
	parentPtr = tablePtr->geomPtr;

	if (argc == 3) {
	    for (childPtr = parentPtr->childPtr; childPtr != NULL;
		     childPtr = childPtr->nextPtr) {
		Tcl_AppendResult(interp, prefix, Tk_PathName(childPtr->tkwin),
				 (char *) NULL);
		prefix = " ";
	    }
	    return TCL_OK;
	}
	return TableAfter(interp, parentPtr, argc-3, argv+3);
    } else if ((c == 'u') && (strncmp(argv[1], "unmanage", length) == 0)) {
	Tk_Window tkwin2;
	Table *tablePtr;
	TkGeomManager *geomPtr;

	if (argc != 3) {
	    Tcl_AppendResult(interp, "wrong # args: should be \"",
		    argv[0], " unmanage window\"", (char *) NULL);
	    return TCL_ERROR;
	}
	tkwin2 = Tk_NameToWindow(interp, argv[2], tkwin);
	if (tkwin2 == NULL) {
	    return TCL_ERROR;
	}
	tablePtr = GetTable(interp, tkwin2);
	geomPtr = tablePtr->geomPtr;
	if (geomPtr->parentPtr != NULL) {
	    Tk_ManageGeometry(tkwin2, (Tk_GeometryProc *) NULL,
		    (ClientData) NULL);
	    TkUnlinkGeomManager(tablePtr);
	    Tk_UnmapWindow(geomPtr->tkwin);
	}
    } else {
	Tcl_AppendResult(interp, "bad option \"", argv[1],
	    "\":  must be configure, manage, or unmanage",
			 (char *) NULL);
	return TCL_ERROR;
    }
    return TCL_OK;
}

/*
 *--------------------------------------------------------------
 *
 * ArrangeTable --
 *
 *	This procedure is invoked (using the Tk_DoWhenIdle
 *	mechanism) to re-layout the window currently shown by
 *	the table.  It is invoked at idle time so that a
 *	series of table requests can be merged into a single
 *	operation.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	The child windows will be rearranged.
 *
 *--------------------------------------------------------------
 */

static void
ArrangeTable(clientData)
    ClientData clientData;	/* Structure describing parent
				 * whose children are to be
				 * re-layed out. */
{
    register Table *tablePtr = (Table *) clientData;
    register TkGeomManager *parentPtr = tablePtr->geomPtr;
    register TkGeomManager *childPtr;
    int row, col;
    int width, height;
    int intBWidth;
    int childNumber;
    int maxChildWidth;
    int maxChildHeight;

    parentPtr->flags &= ~REQUESTED_REARRANGE;

    childNumber = 0;
    maxChildWidth = 0;
    maxChildHeight = 0;

    for (childPtr = parentPtr->childPtr;
	 	childPtr != NULL;
	 	childPtr = childPtr->nextPtr) {
	childNumber++;
	if (Tk_ReqWidth(childPtr->tkwin) > maxChildWidth) {
	    maxChildWidth  = Tk_ReqWidth(childPtr->tkwin);
	}
	if (Tk_ReqHeight(childPtr->tkwin) > maxChildHeight) {
	    maxChildHeight = Tk_ReqHeight(childPtr->tkwin);
	}
    }	

    Tk_Preserve(parentPtr);

    intBWidth = Tk_InternalBorderWidth(parentPtr->tkwin);

    /* 
     * Calculate our (future) size and request it if needed.
     */

    if (tablePtr->arrange_square) {
	int sqr;

	sqr = (int)floor(sqrt((double)childNumber) + 0.99);

	if (tablePtr->arrange_vertical) {
	    tablePtr->rowNumber = sqr;
	} else {
	    tablePtr->columnNumber = sqr;
	}
    }

    if (tablePtr->arrange_vertical) {
	if (tablePtr->rowNumber == 0) {
	    tablePtr->rowNumber = 1;
	}
	tablePtr->columnNumber =
	  (childNumber / tablePtr->rowNumber)
	    + (childNumber % tablePtr->rowNumber != 0); 
    } else {
	if (tablePtr->columnNumber == 0) {
	    tablePtr->columnNumber = 1;
	}
	tablePtr->rowNumber =
	  (childNumber / tablePtr->columnNumber)
	    + (childNumber % tablePtr->columnNumber != 0);
    }
    width  = tablePtr->columnNumber * maxChildWidth + 2*intBWidth;
    height = tablePtr->rowNumber    * maxChildHeight+ 2*intBWidth;

    if ((width != Tk_ReqWidth(parentPtr->tkwin))
	 || (height != Tk_ReqHeight(parentPtr->tkwin))) {
	Tk_GeometryRequest(parentPtr->tkwin, width, height);
	parentPtr->flags |= REQUESTED_REARRANGE;
	Tk_DoWhenIdle(parentPtr->arrangeProc, parentPtr->clientData);
	Tk_Release((ClientData) parentPtr);
	return;
    }

    if (childNumber == 0) {
	Tk_Release((ClientData) parentPtr);
	return;
    }
    /*
     * Calculate granted row and column sizes
     */

    tablePtr->columnWidth
      = (Tk_Width(parentPtr->tkwin) - 2*intBWidth)  / tablePtr->columnNumber;
    tablePtr->rowHeight
      = (Tk_Height(parentPtr->tkwin) - 2*intBWidth) / tablePtr->rowNumber;

    /*
     * Place and map all tabulated children
     */

    row = col = 0;
    for (childPtr = parentPtr->childPtr;
	 childPtr != NULL;
	 childPtr = childPtr->nextPtr) {
	Table *childTablePtr;

	childTablePtr = (Table *)childPtr->clientData;
	childTablePtr->row = row;
	childTablePtr->col = col;

	PlaceChild(childTablePtr, tablePtr);

	if (tablePtr->arrange_vertical) {
	    row++;
	    if (row >= tablePtr->rowNumber) {
		row = 0;
		col++;
	    }
	} else {
	    col++;
	    if (col >= tablePtr->columnNumber) {
		col = 0;
		row++;
	    }
	}
    }
    Tk_Release((ClientData) parentPtr);
}

/*
 *----------------------------------------------------------------------
 *
 * PlaceChild --
 *
 *	This procedure calculates new size and location
 *	for the given child window, and the relocates it if needed.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	The child size and/or location may change.
 */

static void
PlaceChild(childPtr, parentPtr)
    Table *childPtr;
    Table *parentPtr;
{
    int x, y;
    int width, height;
    Tk_Window child = childPtr->geomPtr->tkwin;
    Tk_Window parent = parentPtr->geomPtr->tkwin;
    int intBWidth = Tk_InternalBorderWidth(parent);

    /*
     * Calculate the child size 
     */
    
    if ((childPtr->fillx)
	|| (Tk_ReqWidth(child) > parentPtr->columnWidth)) {
	width = parentPtr->columnWidth;
    } else {
	width = Tk_ReqWidth(child);
    }
    if ((childPtr->filly) 
	|| (Tk_ReqHeight(child) > parentPtr->rowHeight)) {
	height = parentPtr->rowHeight;
    } else {
	height = Tk_ReqHeight(child);
    }
    
    /*
     * Calculate the child location
     */
    
    x = childPtr->col * parentPtr->columnWidth + intBWidth;
    y = childPtr->row * parentPtr->rowHeight + intBWidth;  
    
    switch (childPtr->anchorPoint) {
      case TK_ANCHOR_N:
	x += (parentPtr->columnWidth - width)/2;
	break;
      case TK_ANCHOR_NE:
	x += parentPtr->columnWidth - width;
	break;
      case TK_ANCHOR_E:
	x += parentPtr->columnWidth - width;
	y += (parentPtr->rowHeight - height)/2;
	break;
      case TK_ANCHOR_SE:
	x += parentPtr->columnWidth - width;
	y += parentPtr->rowHeight - height;
	break;
      case TK_ANCHOR_S:
	x += (parentPtr->columnWidth - width)/2;
	y += parentPtr->rowHeight - height;
	break;
      case TK_ANCHOR_SW:
	y += parentPtr->rowHeight - height;
	break;
      case TK_ANCHOR_W:
	y += (parentPtr->rowHeight - height)/2;
	break;
      case TK_ANCHOR_NW:
	break;
      case TK_ANCHOR_CENTER:
	x += (parentPtr->columnWidth - width)/2;
	y += (parentPtr->rowHeight - height)/2;
	break;
      default:
	panic("bad frame factor in ArrangeTable");
    }
    
    /*
     * If the window is too small to be interesting then
     * unmap it.  Otherwise configure it and then make sure
     * it's mapped.
     */
    
    if ((width <= 0) || (height <= 0)) {
	Tk_UnmapWindow(child);
    } else {
	if ((x != Tk_X(child))
	    || (y != Tk_Y(child))
	    || (width != Tk_Width(child))
	    || (height != Tk_Height(child))) {
	    Tk_MoveResizeWindow(child, x, y,
				(unsigned int) width, (unsigned int) height);
	}
	Tk_MapWindow(child);
    }
    
    }


/*
 *--------------------------------------------------------------
 *
 * GetTable --
 *
 *	This internal procedure is used to locate a Table
 *	structure for a given window, creating one if one
 *	doesn't exist already.
 *
 * Results:
 *	The return value is a pointer to the Table structure
 *	corresponding to tkwin.
 *
 * Side effects:
 *	A new table structure may be created.  If so, then
 *	a callback is set up to clean things up when the
 *	window is deleted.
 *
 *--------------------------------------------------------------
 */

static Table *
GetTable(interp, tkwin)
    Tcl_Interp *interp;
    Tk_Window tkwin;		/* Token for window for which
				 * table structure is desired. */
{
    register Table *tablePtr;
    Tcl_HashEntry *hPtr;
    int new;

    if (!initialized) {
	initialized = 1;
	Tcl_InitHashTable(&tableHashTable, TCL_ONE_WORD_KEYS);
    }

    /*
     * See if there's already Table for this window.  If not,
     * then create a new one.
     */

    hPtr = Tcl_CreateHashEntry(&tableHashTable, (char *) tkwin, &new);
    if (!new) {
	return (Table *) Tcl_GetHashValue(hPtr);
    }

    tablePtr = (Table *) ckalloc(sizeof(Table));
    tablePtr->geomPtr = TkCreateGeomManager(interp, tkwin, "table",
			   ArrangeTable, DestroyTable,
			   (ClientData)tablePtr);
    tablePtr->columnNumber = 0;
    tablePtr->rowNumber = 0;
    tablePtr->columnWidth = 0;
    tablePtr->rowHeight = 0;
    tablePtr->anchorPoint = TK_ANCHOR_W;
    tablePtr->row = tablePtr->col = 0;
    tablePtr->fillx = tablePtr->filly = 0;
    tablePtr->arrange_vertical = tablePtr->arrange_square = 0;
    Tcl_SetHashValue(hPtr, tablePtr);
	
    Tk_ConfigureWidget(interp, tablePtr->geomPtr, configSpecs,
		       0, NULL, tablePtr, 0);

    return tablePtr;
}

/*
 *--------------------------------------------------------------
 *
 * TableAfter --
 *
 *	This procedure does most of the real work of adding
 *	one or more windows into the shown set of it's parent.
 *
 * Results:
 *	A standard Tcl return value.
 *
 * Side effects:
 *	The geometry of the specified windows may change, both now and
 *	again in the future.
 *
 *--------------------------------------------------------------
 */

static int
TableAfter(interp, parentPtr, argc, argv)
    Tcl_Interp *interp;		/* Interpreter for error reporting. */
    TkGeomManager *parentPtr; 	/* Parent in which to place windows. */
    int argc;			/* Number of elements in argv. */
    char **argv;		/* Array of window names. */
{
    TkGeomManager *prevPtr;
    register Table *tablePtr;
    Tk_Window tkwin;

    /*
     * By default add to the end of window order.
     */
    prevPtr = parentPtr->childPtr;
    if (prevPtr != NULL) {
	while (prevPtr->nextPtr != NULL) {
	    prevPtr = prevPtr->nextPtr;
	}
    }

    /*
     * Iterate over all of the window specifiers.
     */

    for ( ; argc > 0; argc--, argv++, prevPtr = tablePtr->geomPtr) {

	/*
	 * Find the Table for the window to be placed, and make sure
	 * that the window in which it will be placed is its parent.
	 */

	tkwin = Tk_NameToWindow(interp, argv[0], parentPtr->tkwin);
	if (tkwin == NULL) {
	    return TCL_ERROR;
	}
	if (Tk_Parent(tkwin) != parentPtr->tkwin) {
	    Tcl_AppendResult(interp, "tried to table \"",
		    argv[0], "\" in window that isn't its parent",
		    (char *) NULL);
	    return TCL_ERROR;
	}
	tablePtr = GetTable(interp, tkwin);

	/*
	 * Add the Table to the managed list of parent window's
	 * geometry manager.
	 */
	TkLinkGeomManager(parentPtr, tablePtr->geomPtr, prevPtr);
    }
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * DestroyTable --
 *
 *	This procedure is invoked by TkDestroyGeomManager to clean up
 *	the internal structure of a table at a safe time (when no-one
 *	is using it anymore).
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Everything associated with the table is freed up.
 *
 *----------------------------------------------------------------------
 */

static void
DestroyTable(clientData)
    ClientData clientData;		/* Info about tableed window that
					 * is now dead. */
{
    register Table *tablePtr = (Table *) clientData;
    ckfree((char *) tablePtr);
}

